#
# tenant_management.py
#
# This source file is part of the FoundationDB open source project
#
# Copyright 2013-2022 Apple Inc. and the FoundationDB project authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# FoundationDB Python API

"""Documentation for this API can be found at
https://apple.github.io/foundationdb/api-python.html"""

from fdb import impl as _impl

_tenant_map_prefix = b'\xff\xff/management/tenant_map/'

# If the existence_check_marker is an empty list, then check whether the tenant exists. 
# After the check, append an item to the existence_check_marker list so that subsequent
# calls to this function will not perform the existence check.
#
# If the existence_check_marker is a non-empty list, return None.
def _check_tenant_existence(tr, key, existence_check_marker, force_maybe_commited):
    if not existence_check_marker:
        existing_tenant = tr[key].wait()
        existence_check_marker.append(None)
        if force_maybe_commited:
            raise _impl.FDBError(1021) # maybe_committed
        return existing_tenant != None

    return None

# Attempt to create a tenant in the cluster. If existence_check_marker is an empty
# list, then this function will check if the tenant already exists and fail if it does.
# Once the existence check is completed, it will not be done again if this function
# retries. As a result, this function may return successfully if the tenant is created
# by someone else concurrently. This behavior allows the operation to be idempotent with
# respect to retries.
#
# If the existence_check_marker is a non-empty list, then the existence check is skipped.
@_impl.transactional
def _create_tenant_impl(tr, tenant_name, existence_check_marker, force_existence_check_maybe_committed=False):
    tr.options.set_special_key_space_enable_writes()
    key = b'%s%s' % (_tenant_map_prefix, tenant_name)

    if _check_tenant_existence(tr, key, existence_check_marker, force_existence_check_maybe_committed) is True:
        raise _impl.FDBError(2132) # tenant_already_exists

    tr[key] = b''
    
# Attempt to delete a tenant from the cluster. If existence_check_marker is an empty
# list, then this function will check if the tenant already exists and fail if it does
# not. Once the existence check is completed, it will not be done again if this function
# retries. As a result, this function may return successfully if the tenant is deleted
# by someone else concurrently. This behavior allows the operation to be idempotent with
# respect to retries.
#
# If the existence_check_marker is a non-empty list, then the existence check is skipped.
@_impl.transactional
def _delete_tenant_impl(tr, tenant_name, existence_check_marker, force_existence_check_maybe_committed=False):
    tr.options.set_special_key_space_enable_writes()
    key = b'%s%s' % (_tenant_map_prefix, tenant_name)

    if _check_tenant_existence(tr, key, existence_check_marker, force_existence_check_maybe_committed) is False:
        raise _impl.FDBError(2131) # tenant_not_found

    del tr[key]

def create_tenant(db_or_tr, tenant_name):
    tenant_name = _impl.process_tenant_name(tenant_name)

    # Only perform the existence check when run using a database
    # Callers using a transaction are expected to check existence themselves if required
    existence_check_marker = [] if not isinstance(db_or_tr, _impl.TransactionRead) else [None]
    _create_tenant_impl(db_or_tr, tenant_name, existence_check_marker)

def delete_tenant(db_or_tr, tenant_name):
    tenant_name = _impl.process_tenant_name(tenant_name)

    # Only perform the existence check when run using a database
    # Callers using a transaction are expected to check existence themselves if required
    existence_check_marker = [] if not isinstance(db_or_tr, _impl.TransactionRead) else [None]
    _delete_tenant_impl(db_or_tr, tenant_name, existence_check_marker)



#
# directory_impl.py
#
# This source file is part of the FoundationDB open source project
#
# Copyright 2013-2018 Apple Inc. and the FoundationDB project authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# FoundationDB Python API

import random
import struct
import threading

from fdb import impl as _impl
from fdb import six
import fdb.tuple
from .subspace_impl import Subspace


class AllocatorTransactionState:
    def __init__(self):
        self.lock = threading.Lock()


class HighContentionAllocator (object):

    def __init__(self, subspace):
        self.counters = subspace[0]
        self.recent = subspace[1]
        self.lock = threading.Lock()

    @_impl.transactional
    def allocate(self, tr):
        """Returns a byte string that
            1) has never and will never be returned by another call to this
               method on the same subspace
            2) is nearly as short as possible given the above
        """

        # Get transaction-local state
        if not hasattr(tr, "__fdb_directory_layer_hca_state__"):
            with self.lock:
                if not hasattr(tr, "__fdb_directory_layer_hca_state__"):
                    tr.__fdb_directory_layer_hca_state__ = AllocatorTransactionState()

        tr_state = tr.__fdb_directory_layer_hca_state__

        while True:
            [start] = [self.counters.unpack(k)[0] for k, _ in tr.snapshot.get_range(
                self.counters.range().start, self.counters.range().stop, limit=1, reverse=True)] or [0]

            window_advanced = False
            while True:
                with tr_state.lock:
                    if window_advanced:
                        del tr[self.counters: self.counters[start]]
                        tr.options.set_next_write_no_write_conflict_range()
                        del tr[self.recent: self.recent[start]]

                    # Increment the allocation count for the current window
                    tr.add(self.counters[start], struct.pack("<q", 1))
                    count = tr.snapshot[self.counters[start]]

                if count != None:
                    count = struct.unpack("<q", bytes(count))[0]
                else:
                    count = 0

                window = self._window_size(start)
                if count * 2 < window:
                    break

                start += window
                window_advanced = True

            while True:
                # As of the snapshot being read from, the window is less than half
                # full, so this should be expected to take 2 tries.  Under high
                # contention (and when the window advances), there is an additional
                # subsequent risk of conflict for this transaction.
                candidate = random.randrange(start, start + window)

                with tr_state.lock:
                    latest_counter = tr.snapshot.get_range(self.counters.range().start, self.counters.range().stop, limit=1, reverse=True)
                    candidate_value = tr[self.recent[candidate]]
                    tr.options.set_next_write_no_write_conflict_range()
                    tr[self.recent[candidate]] = b''

                latest_counter = [self.counters.unpack(k)[0] for k, _ in latest_counter]
                if len(latest_counter) > 0 and latest_counter[0] > start:
                    break

                if candidate_value == None:
                    tr.add_write_conflict_key(self.recent[candidate])
                    return fdb.tuple.pack((candidate,))

    def _window_size(self, start):
        # Larger window sizes are better for high contention, smaller sizes for
        # keeping the keys small.  But if there are many allocations, the keys
        # can't be too small.  So start small and scale up.  We don't want this
        # to ever get *too* big because we have to store about window_size/2
        # recent items.
        if start < 255:
            return 64
        if start < 65535:
            return 1024
        return 8192


class Directory(object):
    def __init__(self, directory_layer, path=(), layer=b''):
        self._directory_layer = directory_layer
        self._path = path
        self._layer = layer

    @_impl.transactional
    def create_or_open(self, tr, path, layer=None):
        path = self._tuplify_path(path)
        return self._directory_layer.create_or_open(tr, self._partition_subpath(path), layer)

    @_impl.transactional
    def open(self, tr, path, layer=None):
        path = self._tuplify_path(path)
        return self._directory_layer.open(tr, self._partition_subpath(path), layer)

    @_impl.transactional
    def create(self, tr, path, layer=None, prefix=None):
        path = self._tuplify_path(path)
        return self._directory_layer.create(tr, self._partition_subpath(path), layer, prefix)

    @_impl.transactional
    def list(self, tr, path=()):
        path = self._tuplify_path(path)
        return self._directory_layer.list(tr, self._partition_subpath(path))

    @_impl.transactional
    def move(self, tr, old_path, new_path):
        old_path = self._tuplify_path(old_path)
        new_path = self._tuplify_path(new_path)
        return self._directory_layer.move(tr, self._partition_subpath(old_path), self._partition_subpath(new_path))

    @_impl.transactional
    def move_to(self, tr, new_absolute_path):
        directory_layer = self._get_layer_for_path(())
        new_absolute_path = _to_unicode_path(new_absolute_path)
        partition_len = len(directory_layer._path)
        partition_path = new_absolute_path[:partition_len]
        if partition_path != directory_layer._path:
            raise ValueError("Cannot move between partitions.")

        return directory_layer.move(tr, self._path[partition_len:], new_absolute_path[partition_len:])

    @_impl.transactional
    def remove(self, tr, path=()):
        path = self._tuplify_path(path)
        directory_layer = self._get_layer_for_path(path)
        return directory_layer.remove(tr, self._partition_subpath(path, directory_layer))

    @_impl.transactional
    def remove_if_exists(self, tr, path=()):
        path = self._tuplify_path(path)
        directory_layer = self._get_layer_for_path(path)
        return directory_layer.remove_if_exists(tr, self._partition_subpath(path, directory_layer))

    @_impl.transactional
    def exists(self, tr, path=()):
        path = self._tuplify_path(path)
        directory_layer = self._get_layer_for_path(path)
        return directory_layer.exists(tr, self._partition_subpath(path, directory_layer))

    def get_layer(self):
        return self._layer

    def get_path(self):
        return self._path

    def _tuplify_path(self, path):
        if not isinstance(path, tuple):
            path = (path,)
        return path

    def _partition_subpath(self, path, directory_layer=None):
        directory_layer = directory_layer or self._directory_layer
        return self._path[len(directory_layer._path):] + path

    # Called by all functions that could operate on this subspace directly (move_to, remove, remove_if_exists, exists)
    # Subclasses can choose to return a different directory layer to use for the operation if path is in fact ()
    def _get_layer_for_path(self, path):
        return self._directory_layer


class DirectoryLayer(Directory):

    def __init__(self, node_subspace=Subspace(rawPrefix=b'\xfe'), content_subspace=Subspace(), allow_manual_prefixes=False):
        Directory.__init__(self, self)

        # If specified, new automatically allocated prefixes will all fall within content_subspace
        self._content_subspace = content_subspace
        self._node_subspace = node_subspace

        self._allow_manual_prefixes = allow_manual_prefixes

        # The root node is the one whose contents are the node subspace
        self._root_node = self._node_subspace[self._node_subspace.key()]
        self._allocator = HighContentionAllocator(self._root_node[b'hca'])

    @_impl.transactional
    def create_or_open(self, tr, path, layer=None):
        """ Opens the directory with the given path.

        If the directory does not exist, it is created (creating parent
        directories if necessary).

        If layer is specified, it is checked against the layer of an existing
        directory or set as the layer of a new directory.
        """
        return self._create_or_open_internal(tr, path, layer)

    def _create_or_open_internal(self, tr, path, layer=None, prefix=None, allow_create=True, allow_open=True):
        self._check_version(tr, write_access=False)

        if prefix is not None and not self._allow_manual_prefixes:
            if len(self._path) == 0:
                raise ValueError("Cannot specify a prefix unless manual prefixes are enabled.")
            else:
                raise ValueError("Cannot specify a prefix in a partition.")

        path = _to_unicode_path(path)

        if not path:
            # Root directory contains node metadata and so cannot be opened.
            raise ValueError("The root directory cannot be opened.")

        existing_node = self._find(tr, path).prefetch_metadata(tr)
        if existing_node.exists():
            if existing_node.is_in_partition():
                subpath = existing_node.get_partition_subpath()
                return existing_node.get_contents(self)._directory_layer._create_or_open_internal(
                    tr, subpath, layer, prefix, allow_create, allow_open
                )

            if not allow_open:
                raise ValueError("The directory already exists.")

            if layer and existing_node.layer() != layer:
                raise ValueError("The directory was created with an incompatible layer.")

            return existing_node.get_contents(self)

        if not allow_create:
            raise ValueError("The directory does not exist.")

        self._check_version(tr)

        if prefix == None:
            prefix = self._content_subspace.key() + self._allocator.allocate(tr)

            if len(list(tr.get_range_startswith(prefix, limit=1))) > 0:
                raise Exception("The database has keys stored at the prefix chosen by the automatic prefix allocator: %r." % prefix)

            if not self._is_prefix_free(tr.snapshot, prefix):
                raise Exception("The directory layer has manually allocated prefixes that conflict with the automatic prefix allocator.")

        elif not self._is_prefix_free(tr, prefix):
            raise ValueError("The given prefix is already in use.")

        if len(path) > 1:
            parent_node = self._node_with_prefix(self.create_or_open(tr, path[:-1]).key())
        else:
            parent_node = self._root_node
        if not parent_node:
            # print repr(path[:-1])
            raise ValueError("The parent directory doesn't exist.")

        node = self._node_with_prefix(prefix)
        tr[parent_node[self.SUBDIRS][path[-1]]] = prefix
        if not layer:
            layer = b''

        tr[node[b'layer']] = layer

        return self._contents_of_node(node, path, layer)

    @_impl.transactional
    def open(self, tr, path, layer=None):
        """ Opens the directory with the given path.

        An error is raised if the directory does not exist, or if a layer is
        specified and a different layer was specified when the directory was
        created.
        """
        return self._create_or_open_internal(tr, path, layer, allow_create=False)

    @_impl.transactional
    def create(self, tr, path, layer=None, prefix=None):
        """Creates a directory with the given path (creating parent directories
           if necessary).

        An error is raised if the given directory already exists.

        If prefix is specified, the directory is created with the given physical
        prefix; otherwise a prefix is allocated automatically.

        If layer is specified, it is recorded with the directory and will be
        checked by future calls to open.
        """
        return self._create_or_open_internal(tr, path, layer, prefix, allow_open=False)

    @_impl.transactional
    def move_to(self, tr, new_absolute_path):
        raise Exception('The root directory cannot be moved.')

    @_impl.transactional
    def move(self, tr, old_path, new_path):
        """Moves the directory found at `old_path` to `new_path`.

        There is no effect on the physical prefix of the given directory, or on
        clients that already have the directory open.

        An error is raised if the old directory does not exist, a directory
        already exists at `new_path`, or the parent directory of `new_path` does
        not exist.
        """
        self._check_version(tr)

        old_path = _to_unicode_path(old_path)
        new_path = _to_unicode_path(new_path)

        if old_path == new_path[:len(old_path)]:
            raise ValueError("The destination directory cannot be a subdirectory of the source directory.")

        old_node = self._find(tr, old_path).prefetch_metadata(tr)
        new_node = self._find(tr, new_path).prefetch_metadata(tr)

        if not old_node.exists():
            raise ValueError("The source directory does not exist.")

        if old_node.is_in_partition() or new_node.is_in_partition():
            if not old_node.is_in_partition() or not new_node.is_in_partition() or old_node.path != new_node.path:
                raise ValueError("Cannot move between partitions.")

            return new_node.get_contents(self).move(tr, old_node.get_partition_subpath(), new_node.get_partition_subpath())

        if new_node.exists():
            raise ValueError("The destination directory already exists. Remove it first.")

        parent_node = self._find(tr, new_path[:-1])
        if not parent_node.exists():
            raise ValueError("The parent of the destination directory does not exist. Create it first.")
        tr[parent_node.subspace[self.SUBDIRS][new_path[-1]]] = self._node_subspace.unpack(old_node.subspace.key())[0]
        self._remove_from_parent(tr, old_path)
        return self._contents_of_node(old_node.subspace, new_path, old_node.layer())

    @_impl.transactional
    def remove(self, tr, path=()):
        """Removes the directory, its contents, and all subdirectories.
        Throws an exception if the directory does not exist.

        Warning: Clients that have already opened the directory might still
        insert data into its contents after it is removed.
        """
        return self._remove_internal(tr, path, fail_on_nonexistent=True)

    @_impl.transactional
    def remove_if_exists(self, tr, path=()):
        """Removes the directory, its contents, and all subdirectories, if
        it exists. Returns true if the directory existed and false otherwise.

        Warning: Clients that have already opened the directory might still
        insert data into its contents after it is removed.
        """
        return self._remove_internal(tr, path, fail_on_nonexistent=False)

    def _remove_internal(self, tr, path, fail_on_nonexistent):
        self._check_version(tr)

        if not path:
            raise ValueError("The root directory cannot be removed.")

        path = _to_unicode_path(path)
        node = self._find(tr, path).prefetch_metadata(tr)

        if not node.exists():
            if fail_on_nonexistent:
                raise ValueError("The directory does not exist.")
            else:
                return False

        if node.is_in_partition():
            return node.get_contents(self)._directory_layer._remove_internal(tr, node.get_partition_subpath(), fail_on_nonexistent)

        self._remove_recursive(tr, node.subspace)
        self._remove_from_parent(tr, path)
        return True

    @_impl.transactional
    def list(self, tr, path=()):
        """Returns the names of the specified directory's subdirectories as a
        list of strings.
        """
        self._check_version(tr, write_access=False)

        path = _to_unicode_path(path)
        node = self._find(tr, path).prefetch_metadata(tr)
        if not node.exists():
            raise ValueError("The directory does not exist.")

        if node.is_in_partition(include_empty_subpath=True):
            return node.get_contents(self).list(tr, node.get_partition_subpath())

        return [name for name, cnode in self._subdir_names_and_nodes(tr, node.subspace)]

    @_impl.transactional
    def exists(self, tr, path=()):
        """Returns whether or not the specified directory exists."""
        self._check_version(tr, write_access=False)

        path = _to_unicode_path(path)
        node = self._find(tr, path).prefetch_metadata(tr)

        if not node.exists():
            return False

        if node.is_in_partition():
            return node.get_contents(self).exists(tr, node.get_partition_subpath())

        return True

    ########################################
    #  Private methods for implementation  #
    ########################################

    SUBDIRS = 0
    VERSION = (1, 0, 0)

    def _check_version(self, tr, write_access=True):
        version = tr[self._root_node[b'version']]

        if not version.present():
            if write_access:
                self._initialize_directory(tr)

            return

        version = struct.unpack('<III', bytes(version))

        if version[0] > self.VERSION[0]:
            raise Exception("Cannot load directory with version %d.%d.%d using directory layer %d.%d.%d" % (version + self.VERSION))

        if version[1] > self.VERSION[1] and write_access:
            raise Exception("Directory with version %d.%d.%d is read-only when opened using directory layer %d.%d.%d" % (version + self.VERSION))

    def _initialize_directory(self, tr):
        tr[self._root_node[b'version']] = struct.pack('<III', *self.VERSION)

    def _node_containing_key(self, tr, key):
        # Right now this is only used for _is_prefix_free(), but if we add
        # parent pointers to directory nodes, it could also be used to find a
        # path based on a key.
        if key.startswith(self._node_subspace.key()):
            return self._root_node
        for k, v in tr.get_range(self._node_subspace.range(()).start,
                                 self._node_subspace.pack((key,)) + b'\x00',
                                 reverse=True,
                                 limit=1):
            prev_prefix = self._node_subspace.unpack(k)[0]
            if key.startswith(prev_prefix):
                return self._node_with_prefix(prev_prefix)
        return None

    def _node_with_prefix(self, prefix):
        if prefix == None:
            return None
        return self._node_subspace[prefix]

    def _contents_of_node(self, node, path, layer=None):
        prefix = self._node_subspace.unpack(node.key())[0]

        if layer == b'partition':
            return DirectoryPartition(self._path + path, prefix, self)
        else:
            return DirectorySubspace(self._path + path, prefix, self, layer)

    def _find(self, tr, path):
        n = _Node(self._root_node, (), path)
        for i, name in enumerate(path):
            n = _Node(self._node_with_prefix(tr[n.subspace[self.SUBDIRS][name]]), path[:i + 1], path)
            if not n.exists() or n.layer(tr) == b'partition':
                return n
        return n

    def _subdir_names_and_nodes(self, tr, node):
        sd = node[self.SUBDIRS]
        for k, v in tr[sd.range(())]:
            yield sd.unpack(k)[0], self._node_with_prefix(v)

    def _remove_from_parent(self, tr, path):
        parent = self._find(tr, path[:-1])
        del tr[parent.subspace[self.SUBDIRS][path[-1]]]

    def _remove_recursive(self, tr, node):
        for name, sn in self._subdir_names_and_nodes(tr, node):
            self._remove_recursive(tr, sn)
        tr.clear_range_startswith(self._node_subspace.unpack(node.key())[0])
        del tr[node.range(())]

    def _is_prefix_free(self, tr, prefix):
        # Returns true if the given prefix does not "intersect" any currently
        # allocated prefix (including the root node). This means that it neither
        # contains any other prefix nor is contained by any other prefix.
        return prefix and not self._node_containing_key(tr, prefix) \
            and not len(list(tr.get_range(self._node_subspace.pack((prefix,)), self._node_subspace.pack((_impl.strinc(prefix),)), limit=1)))

    def _is_prefix_empty(self, tr, prefix):
        return len(list(tr.get_range(prefix, _impl.strinc(prefix), limit=1))) == 0


def _to_unicode_path(path):
    if isinstance(path, bytes):
        path = six.text_type(path)

    if isinstance(path, six.text_type):
        return (path,)

    if isinstance(path, tuple):
        path = list(path)
        for i, name in enumerate(path):
            if isinstance(name, bytes):
                path[i] = six.text_type(path[i])
            elif not isinstance(name, six.text_type):
                raise ValueError('Invalid path: must be a unicode string or a tuple of unicode strings')

        return tuple(path)

    raise ValueError('Invalid path: must be a unicode string or a tuple of unicode strings')


directory = DirectoryLayer()


class DirectorySubspace(Subspace, Directory):
    # A DirectorySubspace represents the *contents* of a directory, but it also
    # remembers the path with which it was opened and offers convenience methods
    # to operate on the directory at that path.

    def __init__(self, path, prefix, directory_layer=directory, layer=None):
        Subspace.__init__(self, rawPrefix=prefix)
        Directory.__init__(self, directory_layer, path, layer)

    def __repr__(self):
        return 'DirectorySubspace(path=' + repr(self._path) + ', prefix=' + repr(self.rawPrefix) + ')'


class DirectoryPartition(DirectorySubspace):
    def __init__(self, path, prefix, parent_directory_layer):
        directory_layer = DirectoryLayer(Subspace(rawPrefix=prefix + b'\xfe'), Subspace(rawPrefix=prefix))
        directory_layer._path = path
        DirectorySubspace.__init__(self, path, prefix, directory_layer, b'partition')

        self._parent_directory_layer = parent_directory_layer

    def __repr__(self):
        return 'DirectoryPartition(path=' + repr(self._path) + ', prefix=' + repr(self.rawPrefix) + ')'

    def __getitem__(self, name):
        raise Exception('Cannot open subspace in the root of a directory partition.')

    def key(self):
        raise Exception('Cannot get key for the root of a directory partition.')

    def pack(self, t=tuple()):
        raise Exception('Cannot pack keys using the root of a directory partition.')

    def unpack(self, key):
        raise Exception('Cannot unpack keys using the root of a directory partition.')

    def range(self, t=tuple()):
        raise Exception('Cannot get range for the root of a directory partition.')

    def contains(self, key):
        raise Exception('Cannot check whether a key belongs to the root of a directory partition.')

    def as_foundationdb_key(self):
        raise Exception('Cannot use the root of a directory partition as a key.')

    def subspace(self, tuple):
        raise Exception('Cannot open subspace in the root of a directory partition.')

    def _get_layer_for_path(self, path):
        if path == ():
            return self._parent_directory_layer
        else:
            return self._directory_layer


class _Node (object):

    def __init__(self, subspace, path, target_path):
        self.subspace = subspace
        self.path = path
        self.target_path = target_path
        self._layer = None

    def exists(self):
        return self.subspace is not None

    def prefetch_metadata(self, tr):
        if self.exists():
            self.layer(tr)

        return self

    def layer(self, tr=None):
        if tr:
            self._layer = tr[self.subspace[b'layer']]
        elif self._layer is None:
            raise Exception('Layer has not been read')

        return self._layer

    def is_in_partition(self, tr=None, include_empty_subpath=False):
        return self.exists() and self.layer(tr) == b'partition' and (include_empty_subpath or len(self.target_path) > len(self.path))

    def get_partition_subpath(self):
        return self.target_path[len(self.path):]

    def get_contents(self, directory_layer, tr=None):
        return directory_layer._contents_of_node(self.subspace, self.path, self.layer(tr))



#
# subspace_impl.py
#
# This source file is part of the FoundationDB open source project
#
# Copyright 2013-2018 Apple Inc. and the FoundationDB project authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# FoundationDB Python API

import fdb.tuple


class Subspace (object):

    def __init__(self, prefixTuple=tuple(), rawPrefix=b''):
        self.rawPrefix = fdb.tuple.pack(prefixTuple, prefix=rawPrefix)

    def __repr__(self):
        return 'Subspace(rawPrefix=' + repr(self.rawPrefix) + ')'

    def __getitem__(self, name):
        return Subspace((name,), self.rawPrefix)

    def key(self):
        return self.rawPrefix

    def pack(self, t=tuple()):
        return fdb.tuple.pack(t, prefix=self.rawPrefix)

    def pack_with_versionstamp(self, t=tuple()):
        return fdb.tuple.pack_with_versionstamp(t, prefix=self.rawPrefix)

    def unpack(self, key):
        if not self.contains(key):
            raise ValueError('Cannot unpack key that is not in subspace.')

        return fdb.tuple.unpack(key, prefix_len=len(self.rawPrefix))

    def range(self, t=tuple()):
        p = fdb.tuple.range(t)
        return slice(self.rawPrefix + p.start, self.rawPrefix + p.stop)

    def contains(self, key):
        return key.startswith(self.rawPrefix)

    def as_foundationdb_key(self):
        return self.rawPrefix

    def subspace(self, tuple):
        return Subspace(tuple, self.rawPrefix)



#
# impl.py
#
# This source file is part of the FoundationDB open source project
#
# Copyright 2013-2018 Apple Inc. and the FoundationDB project authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# FoundationDB Python API

import ctypes
import ctypes.util
import datetime
import functools
import inspect
import multiprocessing
import os
import platform
import sys
import threading
import traceback

import fdb
from fdb import six
from fdb.tuple import pack, unpack

_network_thread = None
_network_thread_reentrant_lock = threading.RLock()

_open_file = open

_thread_local_storage = threading.local()

import weakref


class _NetworkOptions(object):
    def __init__(self, parent):
        self._parent = parent


class _ErrorPredicates(object):
    def __init__(self, parent):
        self._parent = parent


class _DatabaseOptions(object):
    def __init__(self, db):
        self._parent = weakref.proxy(db)


class _TransactionOptions(object):
    def __init__(self, tr):
        self._parent = weakref.proxy(tr)


from fdb import fdboptions as _opts
import types
import struct


def option_wrap(code):
    def setfunc(self):
        self._parent._set_option(code, None, 0)
    return setfunc


def option_wrap_string(code):
    def setfunc(self, param=None):
        param, length = optionalParamToBytes(param)
        self._parent._set_option(code, param, length)
    return setfunc


def option_wrap_bytes(code):
    def setfunc(self, param=None):
        if param is None:
            self._parent._set_option(code, None, 0)
        elif isinstance(param, bytes):
            self._parent._set_option(code, param, len(param))
        else:
            raise TypeError('Value must be of type ' + bytes.__name__)
    return setfunc


def option_wrap_int(code):
    def setfunc(self, param):
        self._parent._set_option(code, struct.pack("<q", param), 8)
    return setfunc


def pred_wrap(code):
    def predfunc(self, error):
        return self._parent._error_predicate(code, error.code)
    return predfunc


def operation_wrap(code):
    def opfunc(self, key, param):
        self._atomic_operation(code, key, param)
    return opfunc


def fill_options(scope, predicates=False):
    _dict = getattr(_opts, scope)

    for k, v in _dict.items():
        fname = (predicates and 'is_' or 'set_') + k.lower()
        code, desc, paramType, paramDesc = v
        if predicates:
            f = pred_wrap(code)
        else:
            if paramType == type(None):
                f = option_wrap(code)
            elif paramType == type(''):
                f = option_wrap_string(code)
            elif paramType == type(b''):  # This won't happen in Python 2 because type('') == type(b''), but it will happen in Python 3
                f = option_wrap_bytes(code)
            elif paramType == type(0):
                f = option_wrap_int(code)
            else:
                raise TypeError("Don't know how to set options of type %s" % paramType.__name__)
        f.__name__ = fname
        f.__doc__ = desc
        if paramDesc is not None:
            f.__doc__ += '\n\nArgument is ' + paramDesc
        klass = globals()['_' + scope + 's']
        setattr(klass, fname, f)


def add_operation(fname, v):
    code, desc, paramType, paramDesc = v
    f = operation_wrap(code)
    f.__name__ = fname
    f.__doc__ = desc + '\n\nArguments are the key to which the operation is applied and the ' + paramDesc
    setattr(globals()['Database'], fname, f)
    setattr(globals()['Transaction'], fname, f)


def fill_operations():
    _dict = getattr(_opts, 'MutationType')

    for k, v in _dict.items():
        fname = k.lower()
        add_operation(fname, v)
        add_operation("bit_" + fname, v)


for scope in ['DatabaseOption', 'TransactionOption', 'NetworkOption']:
    fill_options(scope)

fill_options('ErrorPredicate', True)

options = _NetworkOptions(sys.modules[__name__])
predicates = _ErrorPredicates(sys.modules[__name__])


def _set_option(option, param, length):
    _capi.fdb_network_set_option(option, param, length)


def _error_predicate(predicate, error_code):
    return bool(_capi.fdb_error_predicate(predicate, error_code))


def make_enum(scope):
    _dict = getattr(_opts, scope)

    x = type(scope, (), {})

    def makeprop(value, doc):
        return property(fget=lambda o: value, doc=doc)
    for k, v in _dict.items():
        setattr(x, k.lower(), makeprop(v[0], v[1]))

    globals()[scope] = x()


make_enum("StreamingMode")
make_enum("ConflictRangeType")


def transactional(*tr_args, **tr_kwargs):
    """Decorate a funcation as transactional.

    The decorator looks for a named argument (default 'tr') and takes
    one of two actions, depending on the type of the parameter passed
    to the function at call time.

    If given a Database or Tenant, a Transaction will be created and 
    passed into the wrapped code in place of the Database or Tenant. 
    After the function is complete, the newly created transaction 
    will be committed.

    It is important to note that the wrapped method may be called
    multiple times in the event of a commit failure, until the commit
    succeeds.  This restriction requires that the wrapped function
    may not be a generator, or a function that returns a closure that
    contains the `tr` object.

    If given a Transaction, the Transaction will be passed into the
    wrapped code, and WILL NOT be committed at completion of the
    function. This allows new transactional functions to be composed
    of other transactional methods.

    The transactional decorator may be used with or without
    arguments. The keyword argument 'parameter' is recognized and
    specifies the name of the parameter to the wrapped function that
    will contain either a Database or Transaction.

    """
    def decorate(func):
        try:
            parameter = tr_kwargs['parameter']
        except:
            parameter = 'tr'

        wfunc = func
        while getattr(wfunc, '__wrapped__', None):
            wfunc = wfunc.__wrapped__
        if hasattr(inspect, 'getfullargspec'):
            index = inspect.getfullargspec(wfunc).args.index(parameter)
        else:
            index = inspect.getargspec(wfunc).args.index(parameter)

        if getattr(func, '_is_coroutine', False):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                if isinstance(args[index], TransactionRead):
                    raise asyncio.Return((yield asyncio.From(func(*args, **kwargs))))

                largs = list(args)
                tr = largs[index] = args[index].create_transaction()

                while True:
                    try:
                        ret = yield asyncio.From(func(*largs, **kwargs))
                        yield asyncio.From(tr.commit())
                        raise asyncio.Return(ret)
                    except FDBError as e:
                        yield asyncio.From(tr.on_error(e.code))
        else:
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                # We can't throw this from the decorator, as when a user runs
                # >>> import fdb ; fdb.api_version(710)
                # the code above uses @transactional before the API version is set
                if fdb.get_api_version() >= 630 and inspect.isgeneratorfunction(func):
                    raise ValueError("Generators can not be wrapped with fdb.transactional")

                if isinstance(args[index], TransactionRead):
                    return func(*args, **kwargs)

                largs = list(args)
                tr = largs[index] = args[index].create_transaction()

                committed = False
                # retries = 0
                # start = datetime.datetime.now()
                # last = start

                while not committed:
                    ret = None
                    try:
                        ret = func(*largs, **kwargs)
                        if fdb.get_api_version() >= 630 and inspect.isgenerator(ret):
                            raise ValueError("Generators can not be wrapped with fdb.transactional")
                        tr.commit().wait()
                        committed = True
                    except FDBError as e:
                        tr.on_error(e.code).wait()

                    # now = datetime.datetime.now()
                    # td = now - last
                    # elapsed = (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / float(10**6)
                    # if elapsed >= 1:
                    #     td = now - start
                    #     print ('fdb WARNING: long transaction (%gs elapsed in transactional function \'%s\' (%d retries, %s))'
                    #            % (elapsed, func.__name__, retries, committed and 'committed' or 'not yet committed'))
                    #     last = now

                    # retries += 1
                return ret
        return wrapper

    if not tr_args:
        # Being called with parameters (possibly none); return a
        # decorator
        return decorate
    elif len(tr_args) == 1 and not tr_kwargs:
        # Being called as a decorator
        return decorate(tr_args[0])
    else:
        raise Exception('Invalid use of transactional decorator.')


class FDBError(Exception):
    """This exception is raised when an FDB API call returns an
    error. The error code will be stored in the code attribute, and a
    textual description of the error will be stored in the description
    attribute.

    """

    def __init__(self, code):
        self.code = code
        self._description = None

    @property
    def description(self):
        if not self._description:
            self._description = _capi.fdb_get_error(self.code)
        return self._description

    def __str__(self):
        return '%s (%d)' % (self.description, self.code)

    def __repr__(self):
        return 'FDBError(%d)' % self.code


class _FDBBase(object):
    # By inheriting from _FDBBase, every class gets access to self.capi
    # (set below when we open libfdb_capi)
    pass


class FDBRange(object):
    """Iterates over the results of an FDB range query. Returns
    KeyValue objects.

    """

    def __init__(self, tr, begin, end, limit, reverse, streaming_mode):
        self._tr = tr

        self._bsel = begin
        self._esel = end

        self._limit = limit
        self._reverse = reverse
        self._mode = streaming_mode

        self._future = self._tr._get_range(begin, end, limit, streaming_mode, 1, reverse)

    def to_list(self):
        if self._mode == StreamingMode.iterator:
            if self._limit > 0:
                mode = StreamingMode.exact
            else:
                mode = StreamingMode.want_all
        else:
            mode = self._mode

        return list(self.__iter__(mode=mode))

    def __iter__(self, mode=None):
        if mode is None:
            mode = self._mode
        bsel = self._bsel
        esel = self._esel
        limit = self._limit

        iteration = 1  # the first read was fired off when the FDBRange was initialized
        future = self._future

        done = False

        while not done:
            if future:
                (kvs, count, more) = future.wait()
                index = 0
                future = None

                if not count:
                    return

            result = kvs[index]
            index += 1

            if index == count:
                if not more or limit == count:
                    done = True
                else:
                    iteration += 1
                    if limit > 0:
                        limit = limit - count
                    if self._reverse:
                        esel = KeySelector.first_greater_or_equal(kvs[-1].key)
                    else:
                        bsel = KeySelector.first_greater_than(kvs[-1].key)
                    future = self._tr._get_range(bsel, esel, limit, mode, iteration, self._reverse)

            yield result


class TransactionRead(_FDBBase):
    def __init__(self, tpointer, db, snapshot):
        self.tpointer = tpointer
        self.db = db
        self._snapshot = snapshot

    def __del__(self):
        # print('Destroying transactionread 0x%x' % self.tpointer)
        self.capi.fdb_transaction_destroy(self.tpointer)

    def get_read_version(self):
        """Get the read version of the transaction."""
        return FutureInt64(self.capi.fdb_transaction_get_read_version(self.tpointer))

    def get(self, key):
        key = keyToBytes(key)
        return Value(self.capi.fdb_transaction_get(self.tpointer, key, len(key), self._snapshot))

    def get_key(self, key_selector):
        key = keyToBytes(key_selector.key)

        return Key(self.capi.fdb_transaction_get_key(
            self.tpointer, key, len(key), key_selector.or_equal, key_selector.offset, self._snapshot))

    def _get_range(self, begin, end, limit, streaming_mode, iteration, reverse):
        beginKey = keyToBytes(begin.key)
        endKey = keyToBytes(end.key)

        return FutureKeyValueArray(
            self.capi.fdb_transaction_get_range(
                self.tpointer, beginKey, len(beginKey), begin.or_equal, begin.offset,
                endKey, len(endKey), end.or_equal, end.offset,
                limit, 0, streaming_mode, iteration, self._snapshot, reverse))

    def _to_selector(self, key_or_selector):
        if not isinstance(key_or_selector, KeySelector):
            key_or_selector = KeySelector.first_greater_or_equal(key_or_selector)
        return key_or_selector

    def get_range(self, begin, end, limit=0, reverse=False, streaming_mode=StreamingMode.iterator):
        if begin is None:
            begin = b''
        if end is None:
            end = b'\xff'
        begin = self._to_selector(begin)
        end = self._to_selector(end)
        return FDBRange(self, begin, end, limit, reverse, streaming_mode)

    def get_range_startswith(self, prefix, *args, **kwargs):
        prefix = keyToBytes(prefix)
        return self.get_range(prefix, strinc(prefix), *args, **kwargs)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return self.get_range(key.start, key.stop, reverse=(key.step == -1))
        return self.get(key)
    
    def get_estimated_range_size_bytes(self, begin_key, end_key):
        if begin_key is None or end_key is None:
            if fdb.get_api_version() >= 700:
                raise Exception('Invalid begin key or end key')
            else:
                if begin_key is None:
                    begin_key = b''
                if end_key is None:
                    end_key = b'\xff'
        return FutureInt64(self.capi.fdb_transaction_get_estimated_range_size_bytes(
            self.tpointer,
            begin_key, len(begin_key),
            end_key, len(end_key)
            ))
    
    def get_range_split_points(self, begin_key, end_key, chunk_size):
        if begin_key is None or end_key is None or chunk_size <=0:
            raise Exception('Invalid begin key, end key or chunk size')
        return FutureKeyArray(self.capi.fdb_transaction_get_range_split_points(
            self.tpointer,
            begin_key, len(begin_key),
            end_key, len(end_key),
            chunk_size
            ))

class Transaction(TransactionRead):
    """A modifiable snapshot of a Database.

    """

    def __init__(self, tpointer, db):
        super(Transaction, self).__init__(tpointer, db, False)
        self.options = _TransactionOptions(self)
        self.__snapshot = self.snapshot = TransactionRead(tpointer, db, True)

    def __del__(self):
        pass

    def set_read_version(self, version):
        """Set the read version of the transaction."""
        self.capi.fdb_transaction_set_read_version(self.tpointer, version)

    def _set_option(self, option, param, length):
        self.capi.fdb_transaction_set_option(self.tpointer, option, param, length)

    def _atomic_operation(self, opcode, key, param):
        paramBytes = valueToBytes(param)
        paramLength = len(paramBytes)
        keyBytes = keyToBytes(key)
        keyLength = len(keyBytes)
        self.capi.fdb_transaction_atomic_op(self.tpointer, keyBytes, keyLength, paramBytes, paramLength, opcode)

    def set(self, key, value):
        key = keyToBytes(key)
        value = valueToBytes(value)
        self.capi.fdb_transaction_set(self.tpointer, key, len(key), value, len(value))

    def clear(self, key):
        if isinstance(key, KeySelector):
            key = self.get_key(key)

        key = keyToBytes(key)

        self.capi.fdb_transaction_clear(self.tpointer, key, len(key))

    def clear_range(self, begin, end):
        if begin is None:
            begin = b''
        if end is None:
            end = b'\xff'
        if isinstance(begin, KeySelector):
            begin = self.get_key(begin)
        if isinstance(end, KeySelector):
            end = self.get_key(end)

        begin = keyToBytes(begin)
        end = keyToBytes(end)

        self.capi.fdb_transaction_clear_range(self.tpointer, begin, len(begin),
                                              end, len(end))

    def clear_range_startswith(self, prefix):
        prefix = keyToBytes(prefix)
        return self.clear_range(prefix, strinc(prefix))

    def watch(self, key):
        key = keyToBytes(key)
        return FutureVoid(self.capi.fdb_transaction_watch(self.tpointer, key, len(key)))

    def add_read_conflict_range(self, begin, end):
        begin = keyToBytes(begin)
        end = keyToBytes(end)
        self.capi.fdb_transaction_add_conflict_range(self.tpointer, begin, len(begin), end, len(end), ConflictRangeType.read)

    def add_read_conflict_key(self, key):
        key = keyToBytes(key)
        self.add_read_conflict_range(key, key + b'\x00')

    def add_write_conflict_range(self, begin, end):
        begin = keyToBytes(begin)
        end = keyToBytes(end)
        self.capi.fdb_transaction_add_conflict_range(self.tpointer, begin, len(begin), end, len(end), ConflictRangeType.write)

    def add_write_conflict_key(self, key):
        key = keyToBytes(key)
        self.add_write_conflict_range(key, key + b'\x00')

    def commit(self):
        return FutureVoid(self.capi.fdb_transaction_commit(self.tpointer))

    def get_committed_version(self):
        version = ctypes.c_int64()
        self.capi.fdb_transaction_get_committed_version(self.tpointer, ctypes.byref(version))
        return version.value

    def get_approximate_size(self):
        """Get the approximate commit size of the transaction."""
        return FutureInt64(self.capi.fdb_transaction_get_approximate_size(self.tpointer))

    def get_versionstamp(self):
        return Key(self.capi.fdb_transaction_get_versionstamp(self.tpointer))

    def on_error(self, error):
        if isinstance(error, FDBError):
            code = error.code
        elif isinstance(error, int):
            code = error
        else:
            raise error
        return FutureVoid(self.capi.fdb_transaction_on_error(self.tpointer, code))

    def reset(self):
        self.capi.fdb_transaction_reset(self.tpointer)

    def cancel(self):
        self.capi.fdb_transaction_cancel(self.tpointer)

    def __setitem__(self, key, value):
        self.set(key, value)

    def __delitem__(self, key):
        if isinstance(key, slice):
            self.clear_range(key.start, key.stop)
        else:
            self.clear(key)


class Future(_FDBBase):
    Event = threading.Event
    _state = None  # < Hack for trollius

    def __init__(self, fpointer):
        # print('Creating future 0x%x' % fpointer)
        self.fpointer = fpointer

    def __del__(self):
        if self.fpointer:
            # print('Destroying future 0x%x' % self.fpointer)
            self.capi.fdb_future_destroy(self.fpointer)
            self.fpointer = None

    def cancel(self):
        self.capi.fdb_future_cancel(self.fpointer)

    def _release_memory(self):
        self.capi.fdb_future_release_memory(self.fpointer)

    def wait(self):
        raise NotImplementedError

    def is_ready(self):
        return bool(self.capi.fdb_future_is_ready(self.fpointer))

    def block_until_ready(self):
        # Checking readiness is faster than using the callback, so it saves us time if we are already
        # ready. It also doesn't add much to the cost of this function
        if not self.is_ready():
            # Blocking in the native client from the main thread prevents Python from handling signals.
            # To avoid that behavior, we implement the blocking in Python using semaphores and on_ready.
            # Using a Semaphore is faster than an Event, and we create only one per thread to avoid the 
            # cost of creating one every time.
            semaphore = getattr(_thread_local_storage, 'future_block_semaphore', None)
            if semaphore is None:
                semaphore = multiprocessing.Semaphore(0)
                _thread_local_storage.future_block_semaphore = semaphore

            self.on_ready(lambda self: semaphore.release())

            try:
                semaphore.acquire()
            except:
                # If this semaphore didn't actually get released, then we need to replace our thread-local
                # copy so that later callers still function correctly
                _thread_local_storage.future_block_semaphore = multiprocessing.Semaphore(0)
                raise

    def on_ready(self, callback):
        def cb_and_delref(ignore):
            _unpin_callback(cbfunc[0])
            del cbfunc[:]
            try:
                callback(self)
            except:
                try:
                    sys.stderr.write("Discarding uncaught exception from user FDB callback:\n")
                    traceback.print_exception(*sys.exc_info(), file=sys.stderr)
                except:
                    pass
        cbfunc = [_CBFUNC(cb_and_delref)]
        del cb_and_delref
        _pin_callback(cbfunc[0])
        self.capi.fdb_future_set_callback(self.fpointer, cbfunc[0], None)

    @staticmethod
    def wait_for_any(*futures):
        """Does not return until at least one of the given futures is ready.
        Returns the index in the parameter list of a ready future."""
        if not futures:
            raise ValueError("wait_for_any requires at least one future")
        d = {}
        ev = futures[0].Event()
        for i, f in enumerate(futures):
            def cb(ignore, i=i):
                if d.setdefault('i', i) == i:
                    ev.set()
            f.on_ready(cb)
        ev.wait()
        return d['i']

    # asyncio future protocol
    def cancelled(self):
        if not self.done():
            return False
        e = self.exception()
        return getattr(e, 'code', 0) == 1101
    done = is_ready

    def result(self):
        if not self.done():
            raise Exception("Future result not available")
        return self.wait()

    def exception(self):
        if not self.done():
            raise Exception("Future result not available")
        try:
            self.wait()
            return None
        except BaseException as e:
            return e

    def add_done_callback(self, fn):
        self.on_ready(lambda f: self.call_soon_threadsafe(fn, f))

    def remove_done_callback(self, fn):
        raise NotImplementedError()


class FutureVoid(Future):
    def wait(self):
        self.block_until_ready()
        self.capi.fdb_future_get_error(self.fpointer)
        return None


class FutureInt64(Future):
    def wait(self):
        self.block_until_ready()
        value = ctypes.c_int64()
        self.capi.fdb_future_get_int64(self.fpointer, ctypes.byref(value))
        return value.value

class FutureUInt64(Future):
    def wait(self):
        self.block_until_ready()
        value = ctypes.c_uint64()
        self.capi.fdb_future_get_uint64(self.fpointer, ctypes.byref(value))
        return value.value

class FutureKeyValueArray(Future):
    def wait(self):
        self.block_until_ready()
        kvs = ctypes.pointer(KeyValueStruct())
        count = ctypes.c_int()
        more = ctypes.c_int()
        self.capi.fdb_future_get_keyvalue_array(self.fpointer, ctypes.byref(kvs), ctypes.byref(count), ctypes.byref(more))
        return ([KeyValue(ctypes.string_at(x.key, x.key_length), ctypes.string_at(x.value, x.value_length))
                for x in kvs[0:count.value]], count.value, more.value)

        # Logically, we should self._release_memory() after extracting the
        # KVs but before returning, but then we would have to store
        # the KVs on the python side and in most cases we are about to
        # destroy the future anyway

class FutureKeyArray(Future):
    def wait(self):
        self.block_until_ready()
        ks = ctypes.pointer(KeyStruct())
        count = ctypes.c_int()
        self.capi.fdb_future_get_key_array(self.fpointer, ctypes.byref(ks), ctypes.byref(count))
        return [ctypes.string_at(x.key, x.key_length) for x in ks[0:count.value]]


class FutureStringArray(Future):
    def wait(self):
        self.block_until_ready()
        strings = ctypes.pointer(ctypes.c_char_p())
        count = ctypes.c_int()
        self.capi.fdb_future_get_string_array(self.fpointer, ctypes.byref(strings), ctypes.byref(count))
        return list(strings[0:count.value])


class replaceable_property(object):
    def __get__(self, obj, cls=None):
        return self.method(obj)

    def __init__(self, method):
        self.method = method


class LazyFuture(Future):
    def __init__(self, *args, **kwargs):
        super(LazyFuture, self).__init__(*args, **kwargs)

    def wait(self):
        self.value
        return self

    def _getter(self):
        raise NotImplementedError

    @replaceable_property
    def value(self):
        self.block_until_ready()

        try:
            self._getter()
            self._release_memory()

        except:
            e = sys.exc_info()
            if not (isinstance(e[1], FDBError) and e[1].code == 1102):  # future_released
                raise

        return self.value


# This is a workaround to avoid a compiler issue as described here:
# http://bugs.python.org/issue12370
_super = super


class FutureString(LazyFuture):
    def __init__(self, *args):
        self._error = None
        _super(FutureString, self).__init__(*args)

    def getclass(self):
        return bytes
    __class__ = property(getclass)

    def as_foundationdb_key(self):
        return self.value

    def as_foundationdb_value(self):
        return self.value

    def __str__(self):
        return self.value.__str__()

    def __bytes__(self):
        return self.value

    def __repr__(self):
        return self.value.__repr__()

    def __add__(self, rhs):
        if isinstance(rhs, FutureString):
            rhs = rhs.value
        return self.value + rhs

    def __radd__(self, lhs):
        if isinstance(lhs, FutureString):
            lhs = lhs.value
        return lhs + self.value

    def __mul__(self, rhs):
        return self.value * rhs

    def __rmul__(self, lhs):
        return lhs * self.value

    def __lt__(self, rhs):
        if isinstance(rhs, FutureString):
            rhs = rhs.value
        return self.value < rhs

    def __le__(self, rhs):
        if isinstance(rhs, FutureString):
            rhs = rhs.value
        return self.value <= rhs

    def __gt__(self, rhs):
        if isinstance(rhs, FutureString):
            rhs = rhs.value
        return self.value > rhs

    def __ge__(self, rhs):
        if isinstance(rhs, FutureString):
            rhs = rhs.value
        return self.value >= rhs

    def __eq__(self, rhs):
        if isinstance(rhs, FutureString):
            rhs = rhs.value
        return self.value == rhs

    def __ne__(self, rhs):
        return not self == rhs

    def __nonzero__(self):
        return bool(self.value)

    def __int__(self):
        return int(self.value)


def makewrapper(func):
    def tmpfunc(self, *args):
        return func(self.value, *args)
    return tmpfunc


for i in dir(bytes):
    if not i.startswith('_') or i in ('__getitem__', '__getslice__', '__hash__', '__len__'):
        setattr(FutureString, i, makewrapper(getattr(bytes, i)))


class Value(FutureString):
    def _getter(self):
        present = ctypes.c_int()
        value = ctypes.pointer(ctypes.c_byte())
        value_length = ctypes.c_int()
        self.capi.fdb_future_get_value(self.fpointer, ctypes.byref(present),
                                       ctypes.byref(value), ctypes.byref(value_length))
        if present.value:
            self.value = ctypes.string_at(value, value_length.value)
        else:
            self.value = None

    def present(self):
        return self.value is not None


class Key(FutureString):
    def _getter(self):
        key = ctypes.pointer(ctypes.c_byte())
        key_length = ctypes.c_int()
        self.capi.fdb_future_get_key(self.fpointer, ctypes.byref(key), ctypes.byref(key_length))
        self.value = ctypes.string_at(key, key_length.value)


class FormerFuture(_FDBBase):
    def wait(self):
        return self

    def is_ready(self):
        return True

    def block_until_ready(self):
        pass

    def on_ready(self, callback):
        try:
            callback(self)
        except:
            try:
                sys.stderr.write("Discarding uncaught exception from user FDB callback:\n")
                traceback.print_exception(*sys.exc_info(), file=sys.stderr)
            except:
                pass

class _TransactionCreator(_FDBBase):
    def get(self, key):
        return _TransactionCreator.__creator_getitem(self, key)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return self.get_range(key.start, key.stop, reverse=(key.step == -1))
        return _TransactionCreator.__creator_getitem(self, key)

    def get_key(self, key_selector):
        return _TransactionCreator.__creator_get_key(self, key_selector)

    def get_range(self, begin, end, limit=0, reverse=False, streaming_mode=StreamingMode.want_all):
        return _TransactionCreator.__creator_get_range(self, begin, end, limit, reverse, streaming_mode)

    def get_range_startswith(self, prefix, *args, **kwargs):
        return _TransactionCreator.__creator_get_range_startswith(self, prefix, *args, **kwargs)

    def set(self, key, value):
        _TransactionCreator.__creator_setitem(self, key, value)

    def __setitem__(self, key, value):
        _TransactionCreator.__creator_setitem(self, key, value)

    def clear(self, key):
        _TransactionCreator.__creator_delitem(self, key)

    def clear_range(self, begin, end):
        _TransactionCreator.__creator_delitem(self, slice(begin, end))

    def __delitem__(self, key_or_slice):
        _TransactionCreator.__creator_delitem(self, key_or_slice)

    def clear_range_startswith(self, prefix):
        _TransactionCreator.__creator_clear_range_startswith(self, prefix)

    def get_and_watch(self, key):
        return _TransactionCreator.__creator_get_and_watch(self, key)

    def set_and_watch(self, key, value):
        return _TransactionCreator.__creator_set_and_watch(self, key, value)

    def clear_and_watch(self, key):
        return _TransactionCreator.__creator_clear_and_watch(self, key)

    def create_transaction(self):
        pass

    def _atomic_operation(self, opcode, key, param):
        _TransactionCreator.__creator_atomic_operation(self, opcode, key, param)

    #### Transaction implementations ####
    @staticmethod
    @transactional
    def __creator_getitem(tr, key):
        return tr[key].value

    @staticmethod
    @transactional
    def __creator_get_key(tr, key_selector):
        return tr.get_key(key_selector).value

    @staticmethod
    @transactional
    def __creator_get_range(tr, begin, end, limit, reverse, streaming_mode):
        return tr.get_range(begin, end, limit, reverse, streaming_mode).to_list()

    @staticmethod
    @transactional
    def __creator_get_range_startswith(tr, prefix, *args, **kwargs):
        return tr.get_range_startswith(prefix, *args, **kwargs).to_list()

    @staticmethod
    @transactional
    def __creator_setitem(tr, key, value):
        tr[key] = value

    @staticmethod
    @transactional
    def __creator_clear_range_startswith(tr, prefix):
        tr.clear_range_startswith(prefix)

    @staticmethod
    @transactional
    def __creator_get_and_watch(tr, key):
        v = tr.get(key)
        return v, tr.watch(key)

    @staticmethod
    @transactional
    def __creator_set_and_watch(tr, key, value):
        tr.set(key, value)
        return tr.watch(key)

    @staticmethod
    @transactional
    def __creator_clear_and_watch(tr, key):
        del tr[key]
        return tr.watch(key)

    @staticmethod
    @transactional
    def __creator_delitem(tr, key_or_slice):
        del tr[key_or_slice]

    @staticmethod
    @transactional
    def __creator_atomic_operation(tr, opcode, key, param):
        tr._atomic_operation(opcode, key, param)

    # Asynchronous transactions
    @staticmethod
    def declare_asynchronous_transactions():
        Return = asyncio.Return
        From = asyncio.From
        coroutine = asyncio.coroutine

        class TransactionCreator:
            @staticmethod
            @transactional
            @coroutine
            def __creator_getitem(tr, key):
                # raise Return(( yield From( tr[key] ) ))
                raise Return(tr[key])
                yield None

            @staticmethod
            @transactional
            @coroutine
            def __creator_get_key(tr, key_selector):
                raise Return(tr.get_key(key_selector))
                yield None

            @staticmethod
            @transactional
            @coroutine
            def __creator_get_range(tr, begin, end, limit, reverse, streaming_mode):
                raise Return((yield From(tr.get_range(begin, end, limit, reverse, streaming_mode).to_list())))

            @staticmethod
            @transactional
            @coroutine
            def __creator_get_range_startswith(tr, prefix, *args, **kwargs):
                raise Return((yield From(tr.get_range_startswith(prefix, *args, **kwargs).to_list())))

            @staticmethod
            @transactional
            @coroutine
            def __creator_setitem(tr, key, value):
                tr[key] = value
                raise Return()
                yield None

            @staticmethod
            @transactional
            @coroutine
            def __creator_clear_range_startswith(tr, prefix):
                tr.clear_range_startswith(prefix)
                raise Return()
                yield None

            @staticmethod
            @transactional
            @coroutine
            def __creator_get_and_watch(tr, key):
                v = tr.get(key)
                raise Return(v, tr.watch(key))
                yield None

            @staticmethod
            @transactional
            @coroutine
            def __creator_set_and_watch(tr, key, value):
                tr.set(key, value)
                raise Return(tr.watch(key))
                yield None

            @staticmethod
            @transactional
            @coroutine
            def __creator_clear_and_watch(tr, key):
                del tr[key]
                raise Return(tr.watch(key))
                yield None

            @staticmethod
            @transactional
            @coroutine
            def __creator_delitem(tr, key_or_slice):
                del tr[key_or_slice]
                raise Return()
                yield None

            @staticmethod
            @transactional
            @coroutine
            def __creator_atomic_operation(tr, opcode, key, param):
                tr._atomic_operation(opcode, key, param)
                raise Return()
                yield None
        return TransactionCreator

def process_tenant_name(name):
    if isinstance(name, tuple):
        return pack(name)
    elif isinstance(name, bytes):
        return name
    else:
        raise TypeError('Tenant name must be of type ' + bytes.__name__ + ' or of type ' + tuple.__name__)

class Database(_TransactionCreator):
    def __init__(self, dpointer):
        self.dpointer = dpointer
        self.options = _DatabaseOptions(self)

    def __del__(self):
        # print('Destroying database 0x%x' % self.dpointer)
        self.capi.fdb_database_destroy(self.dpointer)

    def _set_option(self, option, param, length):
        self.capi.fdb_database_set_option(self.dpointer, option, param, length)

    def open_tenant(self, name):
        tname = process_tenant_name(name)
        pointer = ctypes.c_void_p()
        self.capi.fdb_database_open_tenant(self.dpointer, tname, len(tname), ctypes.byref(pointer))
        return Tenant(pointer.value)

    def create_transaction(self):
        pointer = ctypes.c_void_p()
        self.capi.fdb_database_create_transaction(self.dpointer, ctypes.byref(pointer))
        return Transaction(pointer.value, self)


class Tenant(_TransactionCreator):
    def __init__(self, tpointer):
        self.tpointer = tpointer

    def __del__(self):
        self.capi.fdb_tenant_destroy(self.tpointer)

    def create_transaction(self):
        pointer = ctypes.c_void_p()
        self.capi.fdb_tenant_create_transaction(self.tpointer, ctypes.byref(pointer))
        return Transaction(pointer.value, self)


fill_operations()


class Cluster(_FDBBase):
    def __init__(self, cluster_file):
        self.cluster_file = cluster_file
        self.options = None

    def open_database(self, name):
        if name != b'DB':
            raise FDBError(2013) # invalid_database_name

        return create_database(self.cluster_file)


def create_database(cluster_file=None):
    pointer = ctypes.c_void_p()
    _FDBBase.capi.fdb_create_database(optionalParamToBytes(cluster_file)[0], ctypes.byref(pointer))
    return Database(pointer)

def create_cluster(cluster_file=None):
    return Cluster(cluster_file)


class KeySelector(object):
    def __init__(self, key, or_equal, offset):
        self.key = key
        self.or_equal = or_equal
        self.offset = offset

    def __add__(self, offset):
        return KeySelector(self.key, self.or_equal, self.offset + offset)

    def __sub__(self, offset):
        return KeySelector(self.key, self.or_equal, self.offset - offset)

    @classmethod
    def last_less_than(cls, key):
        return cls(key, False, 0)

    @classmethod
    def last_less_or_equal(cls, key):
        return cls(key, True, 0)

    @classmethod
    def first_greater_than(cls, key):
        return cls(key, True, 1)

    @classmethod
    def first_greater_or_equal(cls, key):
        return cls(key, False, 1)

    def __repr__(self):
        return 'KeySelector(%r, %r, %r)' % (self.key, self.or_equal, self.offset)


class KVIter(object):
    def __init__(self, obj):
        self.obj = obj
        self.index = 0

    def __iter__(self):
        return self

    def next(self):
        self.index += 1
        if self.index == 1:
            return self.obj.key
        elif self.index == 2:
            return self.obj.value
        else:
            raise StopIteration

    def __next__(self):
        return self.next()


class KeyValueStruct(ctypes.Structure):
    _fields_ = [('key', ctypes.POINTER(ctypes.c_byte)),
                ('key_length', ctypes.c_int),
                ('value', ctypes.POINTER(ctypes.c_byte)),
                ('value_length', ctypes.c_int)]
    _pack_ = 4

class KeyStruct(ctypes.Structure):
    _fields_ = [('key', ctypes.POINTER(ctypes.c_byte)),
                ('key_length', ctypes.c_int)]
    _pack_ = 4


class KeyValue(object):
    def __init__(self, key, value):
        self.key = key
        self.value = value

    def __repr__(self):
        return '%s: %s' % (repr(self.key), repr(self.value))

    def __iter__(self):
        return KVIter(self)


def check_error_code(code, func, arguments):
    if code:
        raise FDBError(code)
    return None


if sys.maxsize <= 2**32:
    raise Exception("FoundationDB API requires a 64-bit python interpreter!")
if platform.system() == 'Windows':
    capi_name = 'fdb_c.dll'
elif platform.system() == 'Linux':
    capi_name = 'libfdb_c.so'
elif platform.system() == 'FreeBSD':
    capi_name = 'libfdb_c.so'
elif platform.system() == 'Darwin':
    capi_name = 'libfdb_c.dylib'
elif sys.platform == 'win32':
    capi_name = 'fdb_c.dll'
elif sys.platform.startswith('cygwin'):
    capi_name = 'fdb_c.dll'
elif sys.platform.startswith('linux'):
    capi_name = 'libfdb_c.so'
elif sys.platform == 'darwin':
    capi_name = 'libfdb_c.dylib'
else:
    raise Exception("Platform (%s) %s is not supported by the FoundationDB API!" % (sys.platform, platform.system()))
this_dir = os.path.dirname(__file__)


# Preferred installation: The C API library or a symbolic link to the
#    library should be in the same directory as this module.
# Failing that, a file named $(capi_name).pth should be in the same directory,
#    and a relative path to the library (including filename)
# Failing that, we try to load the C API library without qualification, and
#    the library should be on the platform's dynamic library search path
def read_pth_file():
    pth_file = os.path.join(this_dir, capi_name + '.pth')
    if not os.path.exists(pth_file):
        return None
    pth = _open_file(pth_file, "rt").read().strip()
    if pth[0] != '/':
        pth = os.path.join(this_dir, pth)
    return pth


for pth in [
    lambda: os.path.join(this_dir, capi_name),
    # lambda: os.path.join(this_dir, '../../lib', capi_name),  # For compatibility with existing unix installation process... should be removed
    read_pth_file
]:
    p = pth()
    if p and os.path.exists(p):
        _capi = ctypes.CDLL(os.path.abspath(p))
        break
else:
    try:
        _capi = ctypes.CDLL(capi_name)
    except:
        # The system python on OS X can't find the library installed to /usr/local/lib if SIP is enabled
        # find_library does find the location in /usr/local/lib, so if the above fails fallback to using it
        lib_path = ctypes.util.find_library(capi_name)
        if lib_path is not None:
            try:
                _capi = ctypes.CDLL(lib_path)
            except:
                raise Exception("Unable to locate the FoundationDB API shared library!")
        else:
            raise Exception("Unable to locate the FoundationDB API shared library!")


def keyToBytes(k):
    if hasattr(k, 'as_foundationdb_key'):
        k = k.as_foundationdb_key()
    if not isinstance(k, bytes):
        raise TypeError('Key must be of type ' + bytes.__name__)
    return k


def valueToBytes(v):
    if hasattr(v, 'as_foundationdb_value'):
        v = v.as_foundationdb_value()
    if not isinstance(v, bytes):
        raise TypeError('Value must be of type ' + bytes.__name__)
    return v


def paramToBytes(v):
    if isinstance(v, FutureString):
        v = v.value
    if not isinstance(v, bytes) and hasattr(v, 'encode'):
        v = v.encode('utf8')
    if not isinstance(v, bytes):
        raise TypeError('Parameter must be a string')
    return v


def optionalParamToBytes(v):
    if v is None:
        return (None, 0)
    else:
        v = paramToBytes(v)
        return (v, len(v))


_FDBBase.capi = _capi
_CBFUNC = ctypes.CFUNCTYPE(None, ctypes.c_void_p)

def init_c_api():
    _capi.fdb_select_api_version_impl.argtypes = [ctypes.c_int, ctypes.c_int]
    _capi.fdb_select_api_version_impl.restype = ctypes.c_int

    _capi.fdb_get_error.argtypes = [ctypes.c_int]
    _capi.fdb_get_error.restype = ctypes.c_char_p

    _capi.fdb_error_predicate.argtypes = [ctypes.c_int, ctypes.c_int]
    _capi.fdb_error_predicate.restype = ctypes.c_int

    _capi.fdb_setup_network.argtypes = []
    _capi.fdb_setup_network.restype = ctypes.c_int
    _capi.fdb_setup_network.errcheck = check_error_code

    _capi.fdb_network_set_option.argtypes = [ctypes.c_int, ctypes.c_void_p, ctypes.c_int]
    _capi.fdb_network_set_option.restype = ctypes.c_int
    _capi.fdb_network_set_option.errcheck = check_error_code

    _capi.fdb_run_network.argtypes = []
    _capi.fdb_run_network.restype = ctypes.c_int
    _capi.fdb_run_network.errcheck = check_error_code

    _capi.fdb_stop_network.argtypes = []
    _capi.fdb_stop_network.restype = ctypes.c_int
    _capi.fdb_stop_network.errcheck = check_error_code

    _capi.fdb_future_destroy.argtypes = [ctypes.c_void_p]
    _capi.fdb_future_destroy.restype = None

    _capi.fdb_future_release_memory.argtypes = [ctypes.c_void_p]
    _capi.fdb_future_release_memory.restype = None

    _capi.fdb_future_cancel.argtypes = [ctypes.c_void_p]
    _capi.fdb_future_cancel.restype = None

    _capi.fdb_future_block_until_ready.argtypes = [ctypes.c_void_p]
    _capi.fdb_future_block_until_ready.restype = ctypes.c_int
    _capi.fdb_future_block_until_ready.errcheck = check_error_code

    _capi.fdb_future_is_ready.argtypes = [ctypes.c_void_p]
    _capi.fdb_future_is_ready.restype = ctypes.c_int

    _capi.fdb_future_set_callback.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p]
    _capi.fdb_future_set_callback.restype = int
    _capi.fdb_future_set_callback.errcheck = check_error_code

    _capi.fdb_future_get_error.argtypes = [ctypes.c_void_p]
    _capi.fdb_future_get_error.restype = int
    _capi.fdb_future_get_error.errcheck = check_error_code

    _capi.fdb_future_get_int64.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_int64)]
    _capi.fdb_future_get_int64.restype = ctypes.c_int
    _capi.fdb_future_get_int64.errcheck = check_error_code

    _capi.fdb_future_get_uint64.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_uint64)]
    _capi.fdb_future_get_uint64.restype = ctypes.c_uint
    _capi.fdb_future_get_uint64.errcheck = check_error_code

    _capi.fdb_future_get_key.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.POINTER(ctypes.c_byte)),
                                         ctypes.POINTER(ctypes.c_int)]
    _capi.fdb_future_get_key.restype = ctypes.c_int
    _capi.fdb_future_get_key.errcheck = check_error_code

    _capi.fdb_future_get_value.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_int),
                                           ctypes.POINTER(ctypes.POINTER(ctypes.c_byte)), ctypes.POINTER(ctypes.c_int)]
    _capi.fdb_future_get_value.restype = ctypes.c_int
    _capi.fdb_future_get_value.errcheck = check_error_code

    _capi.fdb_future_get_keyvalue_array.argtypes = [ctypes.c_void_p, ctypes.POINTER(
        ctypes.POINTER(KeyValueStruct)), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int)]
    _capi.fdb_future_get_keyvalue_array.restype = int
    _capi.fdb_future_get_keyvalue_array.errcheck = check_error_code

    _capi.fdb_future_get_key_array.argtypes = [ctypes.c_void_p, ctypes.POINTER(
        ctypes.POINTER(KeyStruct)), ctypes.POINTER(ctypes.c_int)]
    _capi.fdb_future_get_key_array.restype = int
    _capi.fdb_future_get_key_array.errcheck = check_error_code

    _capi.fdb_future_get_string_array.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.POINTER(ctypes.c_char_p)), ctypes.POINTER(ctypes.c_int)]
    _capi.fdb_future_get_string_array.restype = int
    _capi.fdb_future_get_string_array.errcheck = check_error_code

    _capi.fdb_create_database.argtypes = [ctypes.c_char_p, ctypes.POINTER(ctypes.c_void_p)]
    _capi.fdb_create_database.restype = ctypes.c_int
    _capi.fdb_create_database.errcheck = check_error_code

    _capi.fdb_database_destroy.argtypes = [ctypes.c_void_p]
    _capi.fdb_database_destroy.restype = None

    _capi.fdb_database_open_tenant.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.POINTER(ctypes.c_void_p)]
    _capi.fdb_database_open_tenant.restype = ctypes.c_int
    _capi.fdb_database_open_tenant.errcheck = check_error_code

    _capi.fdb_database_create_transaction.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_void_p)]
    _capi.fdb_database_create_transaction.restype = ctypes.c_int
    _capi.fdb_database_create_transaction.errcheck = check_error_code

    _capi.fdb_database_set_option.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p, ctypes.c_int]
    _capi.fdb_database_set_option.restype = ctypes.c_int
    _capi.fdb_database_set_option.errcheck = check_error_code

    _capi.fdb_tenant_destroy.argtypes = [ctypes.c_void_p]
    _capi.fdb_tenant_destroy.restype = None

    _capi.fdb_tenant_create_transaction.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_void_p)]
    _capi.fdb_tenant_create_transaction.restype = ctypes.c_int
    _capi.fdb_tenant_create_transaction.errcheck = check_error_code

    _capi.fdb_transaction_destroy.argtypes = [ctypes.c_void_p]
    _capi.fdb_transaction_destroy.restype = None

    _capi.fdb_transaction_cancel.argtypes = [ctypes.c_void_p]
    _capi.fdb_transaction_cancel.restype = None

    _capi.fdb_transaction_set_read_version.argtypes = [ctypes.c_void_p, ctypes.c_int64]
    _capi.fdb_transaction_set_read_version.restype = None

    _capi.fdb_transaction_get_read_version.argtypes = [ctypes.c_void_p]
    _capi.fdb_transaction_get_read_version.restype = ctypes.c_void_p

    _capi.fdb_transaction_get.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
    _capi.fdb_transaction_get.restype = ctypes.c_void_p

    _capi.fdb_transaction_get_key.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int]
    _capi.fdb_transaction_get_key.restype = ctypes.c_void_p

    _capi.fdb_transaction_get_range.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_void_p,
                                                ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int,
                                                ctypes.c_int, ctypes.c_int]
    _capi.fdb_transaction_get_range.restype = ctypes.c_void_p

    _capi.fdb_transaction_get_estimated_range_size_bytes.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p, ctypes.c_int]
    _capi.fdb_transaction_get_estimated_range_size_bytes.restype = ctypes.c_void_p

    _capi.fdb_transaction_get_range_split_points.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
    _capi.fdb_transaction_get_range_split_points.restype = ctypes.c_void_p

    _capi.fdb_transaction_add_conflict_range.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
    _capi.fdb_transaction_add_conflict_range.restype = ctypes.c_int
    _capi.fdb_transaction_add_conflict_range.errcheck = check_error_code

    _capi.fdb_transaction_get_addresses_for_key.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int]
    _capi.fdb_transaction_get_addresses_for_key.restype = ctypes.c_void_p

    _capi.fdb_transaction_set_option.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p, ctypes.c_int]
    _capi.fdb_transaction_set_option.restype = ctypes.c_int
    _capi.fdb_transaction_set_option.errcheck = check_error_code

    _capi.fdb_transaction_atomic_op.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
    _capi.fdb_transaction_atomic_op.restype = None

    _capi.fdb_transaction_set.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p, ctypes.c_int]
    _capi.fdb_transaction_set.restype = None

    _capi.fdb_transaction_clear.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int]
    _capi.fdb_transaction_clear.restype = None

    _capi.fdb_transaction_clear_range.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p, ctypes.c_int]
    _capi.fdb_transaction_clear_range.restype = None

    _capi.fdb_transaction_watch.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int]
    _capi.fdb_transaction_watch.restype = ctypes.c_void_p

    _capi.fdb_transaction_commit.argtypes = [ctypes.c_void_p]
    _capi.fdb_transaction_commit.restype = ctypes.c_void_p

    _capi.fdb_transaction_get_committed_version.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_int64)]
    _capi.fdb_transaction_get_committed_version.restype = ctypes.c_int
    _capi.fdb_transaction_get_committed_version.errcheck = check_error_code

    _capi.fdb_transaction_get_approximate_size.argtypes = [ctypes.c_void_p]
    _capi.fdb_transaction_get_approximate_size.restype = ctypes.c_void_p

    _capi.fdb_transaction_get_versionstamp.argtypes = [ctypes.c_void_p]
    _capi.fdb_transaction_get_versionstamp.restype = ctypes.c_void_p

    _capi.fdb_transaction_on_error.argtypes = [ctypes.c_void_p, ctypes.c_int]
    _capi.fdb_transaction_on_error.restype = ctypes.c_void_p

    _capi.fdb_transaction_reset.argtypes = [ctypes.c_void_p]
    _capi.fdb_transaction_reset.restype = None

if hasattr(ctypes.pythonapi, 'Py_IncRef'):
    def _pin_callback(cb):
        ctypes.pythonapi.Py_IncRef(ctypes.py_object(cb))

    def _unpin_callback(cb):
        ctypes.pythonapi.Py_DecRef(ctypes.py_object(cb))
else:
    _active_callbacks = set()
    _pin_callback = _active_callbacks.add
    _unpin_callback = _active_callbacks.remove


def init(event_model=None):
    """Initialize the FDB interface.

    Consider using open() as a higher-level interface.

    Keyword arguments:
    event_model -- the event model to support (default None, also 'gevent')

    """
    with _network_thread_reentrant_lock:
        global _network_thread

        # init should only succeed once; if _network_thread is not
        # None, someone has already successfully called init
        if _network_thread:
            raise FDBError(2000)

        try:
            class NetworkThread(threading.Thread):
                def run(self):
                    try:
                        _capi.fdb_run_network()
                    except FDBError as e:
                        sys.stderr.write('Unhandled error in FoundationDB network thread: %s\n' % e)
                    # print('Network stopped')

            _network_thread = NetworkThread()
            _network_thread.daemon = True
            # may not set actual underlying OS thread name
            _network_thread.name = "fdb-network-thread"

            if event_model is not None:
                if event_model == 'gevent':
                    import gevent

                    if gevent.__version__[0] != '0':
                        def nullf():
                            pass

                        class ThreadEvent(object):
                            has_async_ = hasattr(gevent.get_hub().loop, 'async_')
                            def __init__(self):
                                if ThreadEvent.has_async_:
                                    self.gevent_async = gevent.get_hub().loop.async_()
                                else:
                                    self.gevent_async = getattr(gevent.get_hub().loop, 'async')()

                                self.gevent_async.start(nullf)

                            def set(self):
                                self.gevent_async.send()

                            def wait(self):
                                gevent.get_hub().wait(self.gevent_async)
                    else:
                        # gevent 0.x doesn't have async, so use a pipe.  This doesn't work on Windows.
                        if platform.system() == 'Windows':
                            raise Exception("The 'gevent' event_model requires gevent 1.0 on Windows.")

                        import gevent.socket

                        class ThreadEvent(object):
                            def __init__(self):
                                self.pair = os.pipe()

                            def set(self):
                                os.write(self.pair[1], '!')

                            def wait(self):
                                gevent.socket.wait_read(self.pair[0])

                            def __del__(self):
                                os.close(self.pair[0])
                                os.close(self.pair[1])

                    Future.Event = ThreadEvent

                    def _gevent_block_until_ready(self):
                        e = self.Event()

                        def is_ready_cb(future):
                            e.set()
                        self.on_ready(is_ready_cb)
                        e.wait()

                    Future.block_until_ready = _gevent_block_until_ready
                elif event_model == 'debug':
                    import time

                    class DebugEvent(object):
                        def __init__(self):
                            self.ev = threading.Event()

                        def set(self):
                            self.ev.set()

                        def wait(self):
                            while not self.ev.isSet():
                                self.ev.wait(.001)
                    Future.Event = DebugEvent

                    def _debug_block_until_ready(self):
                        while not self.is_ready():
                            time.sleep(.001)
                    Future.block_until_ready = _debug_block_until_ready
                elif event_model == 'asyncio':
                    global asyncio
                    try:
                        import asyncio
                    except ImportError:
                        import trollius as asyncio

                    if isinstance(asyncio.futures._FUTURE_CLASSES, type):
                        asyncio.futures._FUTURE_CLASSES = (asyncio.futures._FUTURE_CLASSES,)
                    asyncio.futures._FUTURE_CLASSES += (Future,)

                    def _do_not_block(self):
                        if not self.is_ready():
                            raise Exception("Future not ready")
                    Future.block_until_ready = _do_not_block
                    Future.call_soon_threadsafe = asyncio.get_event_loop().call_soon_threadsafe
                    Future._loop = asyncio.get_event_loop()

                    def iterate(self):
                        """Usage:
                            fa = tr.get_range(...).iterate()
                            for k,v in (yield From(fa)):
                                print(k,v)
                                yield From(fa)"""
                        def it():
                            yield asyncio.From(self._future)
                            raise asyncio.Return(self)
                        return it()
                    FDBRange.iterate = iterate
                    AT = _TransactionCreator.declare_asynchronous_transactions()
                    for name in dir(AT):
                        if name.startswith("__TransactionCreator__creator_"):
                            setattr(_TransactionCreator, name, getattr(AT, name))

                    def to_list(self):
                        if self._mode == StreamingMode.iterator:
                            if self._limit > 0:
                                mode = StreamingMode.exact
                            else:
                                mode = StreamingMode.want_all
                        else:
                            mode = self._mode
                        yield asyncio.From(self._future)
                        out = []
                        for kv in self.__iter__(mode=mode):
                            out.append(kv)
                            yield asyncio.From(self._future)
                        raise asyncio.Return(out)
                    FDBRange.to_list = to_list
                else:
                    # Hard coded error
                    raise FDBError(2000)

            _capi.fdb_setup_network()

            # Sketchy... the only error returned by fdb_run_network
            # (invoked by _network_thread) is if the network hasn't
            # been setup, so if we get here without exception we know
            # it has been.
            _network_thread.start()
        except:
            # We assigned _network_thread but didn't succeed in init,
            # so clear it out so the next caller has a chance
            _network_thread = None
            raise


def init_v13(local_address, event_model=None):
    return init(event_model)


open_databases = {}

cacheLock = threading.Lock()

def open(cluster_file=None, event_model=None):
    """Opens the given database (or the default database of the cluster indicated
    by the fdb.cluster file in a platform-specific location, if no cluster_file
    or database_name is provided).  Initializes the FDB interface as required."""

    with _network_thread_reentrant_lock:
        if not _network_thread:
            init(event_model=event_model)

    with cacheLock:
        if cluster_file not in open_databases:
            open_databases[cluster_file] = create_database(cluster_file)

        return open_databases[(cluster_file)]
    

def open_v609(cluster_file=None, database_name=b'DB', event_model=None):
    if database_name != b'DB':
        raise FDBError(2013) # invalid_database_name

    return open(cluster_file, event_model)


def open_v13(cluster_id_path, database_name, local_address=None, event_model=None):
    return open_v609(cluster_id_path, database_name, event_model)


import atexit


@atexit.register
def _stop_on_exit():
    if _network_thread:
        _capi.fdb_stop_network()
        _network_thread.join()


def strinc(key):
    key = key.rstrip(b'\xff')
    if len(key) == 0:
        raise ValueError('Key must contain at least one byte not equal to 0xFF.')

    return key[:-1] + six.int2byte(ord(key[-1:]) + 1)



from distutils.core import setup

try:
    with open("README.rst") as f:
        long_desc = f.read()
except:
    long_desc = ""

setup(name="foundationdb",
      version="7.1.15",
      author="FoundationDB",
      author_email="fdb-dist@apple.com",
      description="Python bindings for the FoundationDB database",
      url="https://www.foundationdb.org",
      packages=['fdb'],
      package_data={'fdb': ["fdb/*.py"]},
      long_description=long_desc,
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Apache Software License',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.0',
          'Programming Language :: Python :: 3.1',
          'Programming Language :: Python :: 3.2',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: Implementation :: CPython',
          'Topic :: Database',
          'Topic :: Database :: Front-Ends'
      ]
      )
